/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.impl.gui.main.call;

import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import net.java.sip.communicator.impl.gui.*;
import net.java.sip.communicator.impl.gui.main.contactlist.*;
import net.java.sip.communicator.impl.gui.utils.*;
import net.java.sip.communicator.service.protocol.*;
import net.java.sip.communicator.service.protocol.event.*;
import net.java.sip.communicator.util.*;
import net.java.sip.communicator.util.skin.*;

/**
 * The dialog created when an incoming call is received.
 *
 * @author Yana Stamcheva
 * @author Adam Netocny
 */
public class ReceivedCallDialog
    extends PreCallDialog
    implements  ActionListener,
                CallListener,
                Skinnable
{
    /**
     * The incoming call to render.
     */
    private final Call incomingCall;

    /**
     * Creates a <tt>ReceivedCallDialog</tt> by specifying the associated call.
     *
     * @param call The associated with this dialog incoming call.
     */
    public ReceivedCallDialog(Call call)
    {
        super(GuiActivator.getResources()
            .getSettingsString("service.gui.APPLICATION_NAME")
            + " "
            + GuiActivator.getResources()
                .getI18NString("service.gui.INCOMING_CALL_STATUS")
                    .toLowerCase());

        this.incomingCall = call;

        OperationSetBasicTelephony<?> telephonyOpSet
            = call.getProtocolProvider()
                .getOperationSet(OperationSetBasicTelephony.class);

        telephonyOpSet.addCallListener(this);

        initCallLabel(getCallLabel());
    }

    /**
     * Initializes the label of the received call.
     *
     * @param callLabel The label to initialize.
     */
    private void initCallLabel(final JLabel callLabel)
    {
        Iterator<? extends CallPeer> peersIter = incomingCall.getCallPeers();

        boolean hasMorePeers = false;
        String text = "";

        ImageIcon imageIcon =
            ImageUtils.scaleIconWithinBounds(ImageLoader
                .getImage(ImageLoader.DEFAULT_USER_PHOTO), 40, 45);

        while (peersIter.hasNext())
        {
            final CallPeer peer = peersIter.next();

            // More peers.
            if (peersIter.hasNext())
            {
                text = callLabel.getText()
                    + peer.getDisplayName() + ", ";

                hasMorePeers = true;
            }
            // Only one peer.
            else
            {
                text = callLabel.getText()
                    + peer.getDisplayName()
                    + " "
                    + GuiActivator.getResources()
                        .getI18NString("service.gui.IS_CALLING");

                byte[] image = CallManager.getPeerImage(peer);

                if (image != null && image.length > 0)
                    imageIcon = ImageUtils.getScaledRoundedIcon(image, 50, 50);
                else
                    // Try to find an image in one of the available contact
                    // sources.
                    new Thread(new Runnable()
                    {
                        public void run()
                        {
                            TreeContactList
                                .setSourceContactImage( peer.getAddress(),
                                                        callLabel,
                                                        50, 50);
                        }
                    }).start();
            }
        }

        if (hasMorePeers)
            text += GuiActivator.getResources()
                .getI18NString("service.gui.ARE_CALLING");

        callLabel.setIcon(imageIcon);
        callLabel.setText(text);
    }

    /**
     * When call is remotely ended we close this dialog.
     * @param event the <tt>CallEvent</tt> that has been triggered
     */
    public void callEnded(CallEvent event)
    {
        Call sourceCall = event.getSourceCall();

        if (sourceCall.equals(incomingCall))
        {
            dispose();
        }
    }

    /**
     * Indicates that an incoming call has been received.
     */
    public void incomingCallReceived(CallEvent event) {}

    /**
     * Indicates that an outgoing call has been created.
     */
    public void outgoingCallCreated(CallEvent event) {}

    /**
     * Reloads icons.
     */
    public void loadSkin()
    {
        this.initCallLabel(getCallLabel());
    }

    /**
     * Answers the call when the call button has been pressed.
     */
    @Override
    public void callButtonPressed()
    {
        CallManager.answerCall(incomingCall);
    }

    /**
     * Hangups the call when the call button has been pressed.
     */
    @Override
    public void hangupButtonPressed()
    {
        NotificationManager.stopSound(NotificationManager.INCOMING_CALL);

        CallManager.hangupCall(incomingCall);
    }
}
